/**
 * Any CSS included here will be global. The classic template
 * bundles Infima by default. Infima is a CSS framework designed to
 * work well for content-centric websites.
 */

/* import fonts: DM Mono and DM Sans */
@import url('https://fonts.googleapis.com/css2?family=DM+Mono:ital,wght@0,300;0,400;0,500;1,300;1,400;1,500&family=DM+Sans:ital,opsz,wght@0,9..40,100..1000;1,9..40,100..1000&display=swap');


/* configure tailwindcss */
@tailwind base;
@tailwind components;
@tailwind utilities;



:root {
  --ifm-font-family-base: 'DM Sans';
  --ifm-heading-font-family: 'DM Sans';
  --ifm-font-family-monospace: 'DM Mono', monospace;
  
  --ifm-code-font-size: 95%;
  --docusaurus-highlighted-code-line-bg: rgba(0, 0, 0, 0.1); 
}



.font-mono {
  font-family: "DM Mono", monospace;
  font-style: normal;
}

.font-sans {
  font-family: "DM Sans", sans-serif;
  font-style: normal;
}

button {
  @apply font-sans;
}

#search_input_react {
  @apply font-sans;
}


.header-github-link::before {
  content: '';
  width: 24px;
  height: 24px;
  display: flex;
  background-color: var(--ifm-navbar-link-color);
  mask-image: url("data:image/svg+xml,%3Csvg viewBox='0 0 24 24' xmlns='http://www.w3.org/2000/svg'%3E%3Cpath d='M12 .297c-6.63 0-12 5.373-12 12 0 5.303 3.438 9.8 8.205 11.385.6.113.82-.258.82-.577 0-.285-.01-1.04-.015-2.04-3.338.724-4.042-1.61-4.042-1.61C4.422 18.07 3.633 17.7 3.633 17.7c-1.087-.744.084-.729.084-.729 1.205.084 1.838 1.236 1.838 1.236 1.07 1.835 2.809 1.305 3.495.998.108-.776.417-1.305.76-1.605-2.665-.3-5.466-1.332-5.466-5.93 0-1.31.465-2.38 1.235-3.22-.135-.303-.54-1.523.105-3.176 0 0 1.005-.322 3.3 1.23.96-.267 1.98-.399 3-.405 1.02.006 2.04.138 3 .405 2.28-1.552 3.285-1.23 3.285-1.23.645 1.653.24 2.873.12 3.176.765.84 1.23 1.91 1.23 3.22 0 4.61-2.805 5.625-5.475 5.92.42.36.81 1.096.81 2.22 0 1.606-.015 2.896-.015 3.286 0 .315.21.69.825.57C20.565 22.092 24 17.592 24 12.297c0-6.627-5.373-12-12-12'/%3E%3C/svg%3E");
  transition: background-color var(--ifm-transition-fast)
    var(--ifm-transition-timing-default);
}

.header-github-link:hover::before {
  background-color: var(--ifm-navbar-link-hover-color);
}

/* TOC sidebar header (desktop and mobile) */
.theme-doc-toc-desktop .table-of-contents::before,
.theme-doc-toc-desktop .tableOfContents::before {
  content: "📄 On this page:";
  display: block;
  font-weight: 600;
  margin-bottom: 0.5rem;
  color: var(--ifm-color-content);
}

.theme-doc-toc-mobile .menu__list::before {
  content: "📄 On this page:";
  display: block;
  font-weight: 600;
  margin: 0.5rem 0 0.25rem 0.5rem;
  color: var(--ifm-color-content);
}
